/*
 * Server.java September 2002
 *
 * Copyright (C) 2002, Niall Gallagher <niallg@email.com>
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of 
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public 
 * License along with this program; if not, write to the Free 
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 * MA 02111-1307 USA
 */

package mamba.serve;

import simple.template.velocity.VelocityContainer;
import simple.http.connect.ConnectionFactory;
import simple.http.connect.Connection;
import simple.http.BufferedPipelineFactory;
import simple.http.PipelineFactory;
import simple.http.serve.HandlerFactory;
import simple.http.ProtocolHandler;
import simple.http.serve.FileContext;
import simple.http.serve.Context;
import simple.http.serve.Format;
import simple.template.TemplateEngine;
import simple.template.Container;
import simple.util.FileProperties;
import javax.net.ServerSocketFactory;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import java.security.KeyStore;
import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.PrintStream;
import java.net.ServerSocket;
import java.util.Properties;
import java.io.File;


/**
 * The initialization sequence for the server involves using the
 * configuration details within the System object and creating the
 * various components required to start the server. The sequence 
 * of events to start the server are:
 * <ul> 
 *    <li>
 *    Establish a Context that will be used to acquire files
 *    from service objects.
 *    </li>
 *    <li>
 *    Create a Container that can be used to access template
 *    files used for the display.
 *    </li>
 *    <li>
 *    Create a TemplateEngine that can be used to load service
 *    and controller objects that will use the templating system
 *    to process requests.
 *    </li>
 *    <li>
 *    Create a ProtocolHandler that will perform access logging
 *    and various other server wide tasks.
 *    <li>
 *    <li>
 *    Establish the logging process to be used by the system, 
 *    including redirection of System.out and System.err streams.
 *    <li>
 *    <li>
 *    Create a connection on the specified ports for plain TCP 
 *    connections and possible SSL connections.
 *    <li>
 * </pre>
 * Once all procedures have been completed the server is ready to
 * accept incomming HTTP connections and respond using the 
 * appropriate services. 
 * <p>
 * Before starting the server it must be noted that there are 
 * various properties files that must be edited in order to 
 * configure the server properly. These are:
 * <ul>
 *    <li>mapper.properties</li>
 *    <li>server.properties</li>
 *    <li>velocity.properties</li>
 * </ul>
 * <h2>mapper.properties</h2>
 * <p>
 * The mapper.properties file contains the mapping used to resolve
 * a request URL to a service implementation. There are various 
 * mapping schemes that can be used by the server, such as:
 * <ul>
 *    <li>simple.http.load.PrefixMapper</li>
 *    <li>simple.http.load.PatternMapper</li>
 *    <li>simple.http.load.DefaultMapper</li>
 * </ul>
 * A specific scheme can be specified to the server by using the
 * <code>simple.http.load.mapper</code> property, for example if 
 * mapping was required it simply needs to be specified either on
 * the command line as follows: 
 * <pre>
 *    java -Dsimple.http.load.mapper=&lt;mapper name&gt;
 * </pre>
 * Or it can be specified in the server.properties file. Each of 
 * the mappers has various advantages, and each applies a unique 
 * mapping technique.
 * </p> 
 * <h3>simple.http.load.PrefixMapper</h3>
 * <p>
 * This mapper s used to perform mapping using prefix paths. This
 * provides a scheme like the Servlet context  mapping scheme 
 * where a path prefix within a URI is used to determine which 
 * service is to be given the request. This allows arbitrary path
 * configurations to be mapped to a service class and performs 
 * the mapping task quickly.
 * </p>
 * <h3>simple.http.load.PatternMapper</h3>
 * <p>
 * This is used to perform mapping using patterns. This provides 
 * a scheme like the Servlet wild card mapping scheme where 
 * patterns such as *.html can be used to specify a service 
 * implementation to use. This mapping scheme is very flexible 
 * and also quite fast.
 * </p>
 * <h3>simple.http.load.DefaultMapper</h3>
 * <p>
 * This is used when the service and path used in the transaction
 * need to be explicitly given. This does not require the 
 * mapper.properties file and can adapt to changes in implemented
 * services. For example a service can be used by specifing a URI 
 * such as /com/domain/ExampleService.class/index.html, which 
 * will load and execute com.domain.ExampleService.
 * </p>
 * <h2>server.properties</h2>
 * <p>
 * This is used to directly configure the server. The properties
 * specified in this file will be loaded into the 
 * <code>System</code> object so that they are globaly accessable
 * to objects within the system. The various properties that can
 * be set in this file that are specific to the server are:
 * <table border='1'>
 * <tr>
 *    <td><b>server.properties</b></td>
 *    <td>This specifies where the server.properties file is 
 *    loaded from. If this is not specified then the properties
 *    must be set using the command line.
 *    </td>
 * </tr>
 * <tr>
 *    <td><b>server.port</b></td>
 *    <td>This specifies the port the server listens to for
 *    incomming HTTP requests. This will typically be port 80,
 *    the default for HTTP servers.
 *    </td>
 * </tr>
 * <tr>
 *    <td><b>server.path.class</b></td>
 *    <td>This specifies where the service objects are loaded
 *    from. This should be seperate from the CLASSPATH used so
 *    that the service objects can be reloaded when the byte 
 *    codes change.
 *    </td>
 * </tr>
 * <tr>
 *    <td><b>server.path.log</b></td>
 *    <td>This specifies where the log files are kept for the
 *    server. Typically this will house the access.log, 
 *    error.log, and output.log. The access.log will contain 
 *    details on the client access to the server. The 
 *    error.log and output.log files will contain the output 
 *    from the System.err and System.out streams respectively.
 *    <td>
 * </tr>
 * <tr>
 *    <td><b>server.path.context</b></td>
 *    <td>This specifies where the files that can be accessed
 *    by the service implementations are kept. This acts much 
 *    like the traditional htdocs folder usied by servers like 
 *    Apache. However, not all files will be served from here 
 *    as the templating system can specify a path that it can
 *    use for loading template files.
 *    <p>
 *    By default the templating system used is for the 
 *    Velocity templating system which uses a
 *    velocity.properties file to specify where the template
 *    files are loaded from. This should be the same as the 
 *    server.path.context property for simplicity.
 *    </td>
 * </tr>
 * <tr>  
 *    <td><b>server.ssl.port</b></td> 
 *    <td>This specifies the SSL port to listen to for HTTPS
 *    connections. This will enable content to be encrypted
 *    and secured from interception. This is optional and can
 *    be left unspecified if no HTTPS connections are required.
 *    </td>
 * </tr>
 * <tr>  
 *    <td><b>server.ssl.keystore</b></td> 
 *    <td>This specifies a JKS keystore that is used to store
 *    SSL certificates that will be used to exchange keys for
 *    use in HTTPS communication. The keystore will typically
 *    be created using the keytool application that is 
 *    packaged with the Sun JDK. 
 *    </td>
 * </tr>
 * <tr>  
 *    <td><b>simple.ssl.keystorePassword</b></td> 
 *    <td>To access the JKS keystore and acquire a key a 
 *    keystore password needs to be specified. 
 *    </td>
 * </tr>
 * <tr>  
 *    <td><b>simple.ssl.keyPassword</b></td> 
 *    <td>To access the key within the JKS keystore a key
 *    password needs to be specified.
 *    </td>
 * </tr>
 * </table>
 * </p>
 * <h2>velocity.properties</h2>
 * <p>
 * This is used to configure the
 * <a href='http://jakarta.apache.org/velocity'>Velocity</a> 
 * template engine. Properties in this file influence how template
 * are loaded and processes as well as where the log files for
 * Velocity are kept. Typically there are only two properties that
 * need to be set within this file, they are the following: 
 * <table border='1'>
 * <tr>
 *    <td><b>runtime.log</b></td>
 *    <td>This specifies where the log file for the Velocity
 *    runtime is kept. This should typically refer to a file 
 *    within the server.path.log path so that all log files 
 *    exist within the same place.
 *    </td>
 * </tr>
 * <tr>
 *    <td><b>file.resource.loader.path</b></td>
 *    <td>This specifies where the templates are loaded from by
 *    the Velocity runtime. This should be the same directory 
 *    as the server.path.context so that all resources are 
 *    within the same directory.
 *    </td>
 * </tr>
 * </p>
 * <p>
 * Once all properties files have been edited the server can be
 * started using either the provides bash script for Linux and 
 * UNIX platforms or the following command for Windows:
 * <pre> 
 *    java -Dserver.properties=server.properties \
 *         -classpath &lt;jar files to use&gt; \
 *          mamba.serve.Server
 * </pre>
 * The jar files to use will typically be all the jar files 
 * within the lib directory, however not all  of these jar files
 * are required for the server to start and operate properly.
 *
 * @author Niall Gallagher
 */
public final class Server {

   public static final String SERVER_PROPERTIES = "server.properties";

   public static final String SERVER_PORT = "server.port";

   public static final String SERVER_PATH_CONTEXT = "server.path.context";

   public static final String SERVER_PATH_CLASS = "server.path.class";

   public static final String SERVER_PATH_LOG = "server.path.log";

   public static final String SERVER_SSL_PORT = "server.ssl.port";

   public static final String SERVER_SSL_KEYSTORE = "server.ssl.keystore";

   public static final String SERVER_SSL_KEYSTORE_PASSWORD = "server.ssl.keystorePassword";

   public static final String SERVER_SSL_KEY_PASSWORD = "server.ssl.keyPassword";

   private static ProtocolHandler handler;

   private static Connection connection;

   private static Context context;

   private static Container container;

   private static TemplateEngine engine;

   /**
    * This starts the server using a set of properties that can be
    * taken from the <code>System.getProperty</code> method or by
    * loading the Java properties file specified by the system 
    * property <code>server.properties</code>. To set this
    * property within the VM simply using the following:
    * <pre>
    *  java -Dserver.properties=server.properties
    * </pre>
    * If the above property is set, then all values are pushed in
    * to the system using <code>System.setProperty</code>. This 
    * ensures that they are available throughout the VM.
    */
   public static void main(String list[]) throws Exception {
      Properties conf = System.getProperties();
      String path = System.getProperty(SERVER_PROPERTIES);

      if(path != null){
         conf.putAll(new FileProperties(path));         
      } 
      start();
   }  

   public static void start() throws Exception {
      /*
       * If the "mamba.serve.path" property has been specified a context
       * can be extablished with the directory. This also enables the
       * container to be initialized to server templates.
       */
      Server.context = new FileContext(new File(getServerPath()));
      Server.container = new VelocityContainer();

      /*
       * If the "mamba.serve.classpath" property has been specified the
       * template engine can be initialized to load service objects from the
       * directory. This will used the context path as "mamba.serve.path".
       */
      Server.engine = new TemplateEngine(context, container, new File(getServerClassPath()));

      /*
       * If the "mamba.serve.log" property has been specified a log file
       * can be specified to keep track of the visits made to the site.
       */
      Server.handler = HandlerFactory.getInstance(engine);
      Server.handler = new ServerHandler(Server.handler, new File(getServerLogPath(), "access.log"));

      /*
       * Here the standard error and standard output are redirected to
       * lof files so that they can be examined by an administrator.
       */
      System.setOut(new PrintStream(new FileOutputStream(new File(getServerLogPath(), "output.log"), true), true));
      System.setErr(new PrintStream(new FileOutputStream(new File(getServerLogPath(), "error.log"), true), true));      

      /*
       * When all manditory properties have been specified the server
       * can establish a connection to route incomming HTTP requests to
       * the template engine so that services and templates can be used.
       */
      Server.connection = ConnectionFactory.getConnection(Server.handler, new BufferedPipelineFactory(2048));
      int port = getServerPort();

      if(port > 0){
         Server.connection.connect(new ServerSocket(port));
      }
      int sslPort = getServerSSLPort();

      if(sslPort > 0) {
         String source = getServerKeyStore();
         char[] storePassword = getServerKeyStorePassword().toCharArray();
         char[] keyPassword = getServerKeyPassword().toCharArray();
         KeyStore store = KeyStore.getInstance("JKS");
         store.load(new FileInputStream(source), storePassword);

         KeyManagerFactory keyFactory = 
           KeyManagerFactory.getInstance("SunX509");

         keyFactory.init(store, keyPassword);

         SSLContext sslContext = 
           SSLContext.getInstance("SSLv3");
         sslContext.init(keyFactory.getKeyManagers(), null, null);
         ServerSocketFactory socketFactory = sslContext.getServerSocketFactory();

         /*
          * To establish an SSL connection a key store needs to be specifed
          * and the port to listen for incomming SSL connections will use
          * the keys within the key store to secure the channel.
          */
         Server.connection.connect(socketFactory.createServerSocket(sslPort));
      }
   }

   /**
    * Manditory.
    */
   public static String getServerPath() throws Exception {
      String path = System.getProperty(SERVER_PATH_CONTEXT);

      if(path != null){
         if(!exists(path)) {
            throw new Exception("The "+SERVER_PATH_CONTEXT+" property "+path+
               " refers to an directory that does not exist");
         }         
      } else {
         throw new Exception("The "+SERVER_PATH_CONTEXT+" property must be"+
            " specified as an existing directory");
      }
      return path;
   }


   /**
    * Manditory.
    */
   public static String getServerClassPath() throws Exception {
      String path = System.getProperty(SERVER_PATH_CLASS);

      if(path != null){
         if(!exists(path)) {
            throw new Exception("The "+SERVER_PATH_CLASS+" property "+path+
               " refers to an directory that does not exist");
         }
      } else {
         throw new Exception("The "+SERVER_PATH_CLASS+" property must be"+
            " specified as an existing directory");
      }
      return path;
   }


   /**
    * Manditory.
    */
   public static String getServerLogPath() throws Exception {
      String path = System.getProperty(SERVER_PATH_LOG);

      if(path != null){
         if(!exists(path)) {
            throw new Exception("The "+SERVER_PATH_LOG+" property "+path+
               " refers to an directory that does not exist");
         }
      } else {
         throw new Exception("The "+SERVER_PATH_LOG+" property must be"+
            " specified as an existing directory");
      }
      return path;
   }

   /**
    * Manditory.
    */
   public static int getServerPort() throws Exception {
      String text = System.getProperty(SERVER_PORT);
      int port = -1;

      if(text != null){
         try {
            port = Integer.parseInt(text);
         }catch(NumberFormatException e) {
            throw new Exception("The "+SERVER_PORT+" property must be"+
               " specified as a numeric value");
         }
      } else {
         throw new Exception("The "+SERVER_PORT+" property must be"+
            " specified to start the server");
      }
      return port;
   }

   /**
    * Optional.
    */
   public static int getServerSSLPort() throws Exception{
      String text = System.getProperty(SERVER_SSL_PORT);
      int port = -1;

      if(text != null){
         try {
            port = Integer.parseInt(text);
         }catch(NumberFormatException e) {
            throw new Exception("The "+SERVER_SSL_PORT+" property must be"+
               " specified as a numeric value");
         }
      } else {
         port = -1;   
      }
      return port;
   }  

   public static String getServerKeyStore() throws Exception{
      String store = System.getProperty(SERVER_SSL_KEYSTORE);

      if(store == null){
         throw new Exception("The "+SERVER_SSL_KEYSTORE+" property must be"+
            " specified to enable ssl connections");
      }  
      return store;
   }

   public static String getServerKeyStorePassword() throws Exception{
      String storePassword = System.getProperty(SERVER_SSL_KEYSTORE_PASSWORD);

      if(storePassword == null){
         throw new Exception("The "+SERVER_SSL_KEYSTORE_PASSWORD+" property must be"+
            " specified to enable ssl connections");
      }
      return storePassword;
   }

   public static String getServerKeyPassword() throws Exception{
      String keyPassword = System.getProperty(SERVER_SSL_KEY_PASSWORD);

      if(keyPassword == null){
         throw new Exception("The "+SERVER_SSL_KEY_PASSWORD+" property must be"+
            " specified to enable ssl connections");
      }
      return keyPassword;
   }

   public static boolean exists(String path) {
      return new File(path).exists();
   }
}











